@@ -1,9 +1,10 @@
 import os, socket
 from fasteners import InterProcessLock
-import sys
-import psutil
+import sys, psutil
+from exploitfarm import __version__
 
 DEFAULT_SERVER_PORT = 5050
+DEV_MODE = __version__ == "0.0.0"
 
 def parse_address(address:str, default_port:int|None = None) -> tuple[str, int]:
     parts = address.split(":")
@@ -13,6 +14,10 @@ def parse_address(address:str, default_port:int|None = None) -> tuple[str, int]:
         raise ValueError("Port is required")
     return address, port
 
+def mem_usage() -> float:
+    mem_stats = psutil.virtual_memory()
+    return round((mem_stats.used / mem_stats.total)*100, 1)
+
 def try_tcp_connection(address:str, timeout:float|None = 3) -> tuple[bool, str|None]:
     address, port = parse_address(address, default_port=DEFAULT_SERVER_PORT)
     s = socket.socket()
@@ -44,4 +49,5 @@ def restart_program():
         pass
 
     python = sys.executable
-    os.execl(python, python, *sys.argv)
\ No newline at end of file
+    os.execl(python, python, *sys.argv)
+
